/*
 * seccomp-bpf/hello-raw.c --
 *
 *      Deny all system calls except the exit_group() system call and
 *      the write() system call writing to the standard output.
 */

#define _POSIX_C_SOURCE 200809L

#include <linux/audit.h>
#include <linux/bpf.h>
#include <linux/filter.h>
#include <linux/seccomp.h>
#include <linux/unistd.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/prctl.h>
#include <unistd.h>
#include "util.h"

static void install_filter(void)
{
    int rc;
    struct sock_filter filter[] = {
        BPF_STMT(BPF_LD  | BPF_W | BPF_ABS, (offsetof(struct seccomp_data, arch))),
        BPF_JUMP(BPF_JMP | BPF_JEQ | BPF_K, MYARCH, 0, 7),
        BPF_STMT(BPF_LD  | BPF_W | BPF_ABS, (offsetof(struct seccomp_data, nr))),
        BPF_JUMP(BPF_JMP | BPF_JEQ | BPF_K, __NR_exit_group, 0, 1),
        BPF_STMT(BPF_RET | BPF_K, SECCOMP_RET_ALLOW),
        BPF_JUMP(BPF_JMP | BPF_JEQ | BPF_K, __NR_write, 0, 3),
        BPF_STMT(BPF_LD  | BPF_W | BPF_ABS, (offsetof(struct seccomp_data, args[0]))),
        BPF_JUMP(BPF_JMP | BPF_JEQ | BPF_K, STDOUT_FILENO, 0, 1),
        BPF_STMT(BPF_RET | BPF_K, SECCOMP_RET_ALLOW),
        BPF_STMT(BPF_RET | BPF_K, SECCOMP_RET_KILL),
    };
    struct sock_fprog prog = {
        .len = (unsigned short)(sizeof(filter) / sizeof(filter[0])),
        .filter = filter,
    };

    /* Ensure that execve() never gives the process additional
     * privileges, a necessary prerequisite to install filters.
     * Afterwards, install the BPF filter into the kernel. */
    
    rc = prctl(PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0);
    if (rc < 0) {
        fatal("prctl(NO_NEW_PRIVS, 1, ...) failed\n");
    }
    rc = prctl(PR_SET_SECCOMP, SECCOMP_MODE_FILTER, &prog);
    if (rc < 0) {
        fatal("prctl(PR_SET_SECCOMP, SECCOMP_MODE_FILTER, ...) failed\n");
    }
}

int main(void)
{
    fputs("hello ", stdout);            /* triggers library initializations */
    install_filter();
    fputs("world!\n", stdout);
    fputs("process dying (invalid write system call)\n", stderr);
    return EXIT_SUCCESS;
}
